/*******************************************************************************
 * Copyright (c) 2000, 2003 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * DWT: SWT port for the D Language
 * 
 * Contributors:
 * 	The DWT team at http://www.dsource.org/projects/dwt
 * 
 *******************************************************************************/



module dwt.graphics.fontdata;

/+
import org.eclipse.swt.internal.*;
import org.eclipse.swt.internal.win32.*;
import org.eclipse.swt.*;
+/

private import dwt.dwt;

private import dwt.internal.callback;

private import dwt.internal.win32.os;
private import dwt.internal.win32.tchar;
private import dwt.internal.win32.types;

private import dwt.util.string;
//private import dwt.util.util;

private import std.string;
private import std.utf;


/**
 * Instances of this class describe operating system fonts.
 * Only the public API of this type is platform independent.
 * <p>
 * For platform-independent behaviour, use the get and set methods
 * corresponding to the following properties:
 * <dl>
 * <dt>height</dt><dd>the height of the font in points</dd>
 * <dt>name</dt><dd>the face name of the font, which may include the foundry</dd>
 * <dt>style</dt><dd>A bitwise combination of NORMAL, ITALIC and BOLD</dd>
 * </dl>
 * If extra, platform-dependent functionality is required:
 * <ul>
 * <li>On <em>Windows</em>, the data member of the <code>FontData</code>
 * corresponds to a Windows <code>LOGFONT</code> structure whose fields
 * may be retrieved and modified.</li>
 * <li>On <em>X</em>, the fields of the <code>FontData</code> correspond
 * to the entries in the font's XLFD name and may be retrieved and modified.
 * </ul>
 * Application code does <em>not</em> need to explicitly release the
 * resources managed by each instance when those instances are no longer
 * required, and thus no <code>dispose()</code> method is provided.
 *
 * @see Font
 */

public final class FontData {

	/**
	 * A Win32 LOGFONT struct
	 * (Warning: This field is platform dependent)
	 */
	public LOGFONT* data;

	/**
	 * The height of the font data in points
	 * (Warning: This field is platform dependent)
	 */
	public int height;

	/**
	 * The locales of the font
	 * (Warning: These fields are platform dependent)
	 */
	wchar[] lang, country, variant;

/**
 * Constructs a new un-initialized font data.
 */
public this() {
	data = OS.IsUnicode
            ? cast(LOGFONT*)new LOGFONTW()
            : cast(LOGFONT*)new LOGFONTA();
	// We set the charset field so that
	// wildcard searching will work properly
	// out of the box
	data.lfCharSet = OS.DEFAULT_CHARSET;
	height = 12;
}

/**
 * Constructs a new font data given the Windows <code>LOGFONT</code>
 * that it should represent.
 *
 * @param data the <code>LOGFONT</code> for the result
 */
this(LOGFONT* data, int height) {
	this.data = data;
	this.height = height;
}

/**
 * Constructs a new FontData given a string representation
 * in the form generated by the <code>FontData.toString</code>
 * method.
 * <p>
 * Note that the representation varies between platforms,
 * and a FontData can only be created from a string that was
 * generated on the same platform.
 * </p>
 *
 * @param string the string representation of a <code>FontData</code> (must not be null)
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the argument is null</li>
 *    <li>ERROR_INVALID_ARGUMENT - if the argument does not represent a valid description</li>
 * </ul>
 *
 * @see #toString
 */
public this(wchar[] string) {
	wchar[][] str = Strings!(wchar).split(string, "|");

	if (string === null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
	int pos = 0;
	if (pos >= str.length) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	wchar[] version1 = str[pos++];
	//try {
		if (Strings!(wchar).toInt(version1) != 1) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	//} catch (NumberFormatException e) {
	//	DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	//}

	if (pos >= str.length) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	wchar[] name = str[pos++];

	if (pos >= str.length) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	int height = 0;
	//try {
		height = Strings!(wchar).toInt(str[pos++]);
	//} catch (NumberFormatException e) {
	//	DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	//}

	if (pos >= str.length) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	int style = 0;
	//try {
		style = Strings!(wchar).toInt(str[pos++]);
	//} catch (NumberFormatException e) {
	//	DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	//}

	data = OS.IsUnicode
		? cast(LOGFONT*)new LOGFONTW()
		: cast(LOGFONT*)new LOGFONTA();
	data.lfCharSet = OS.DEFAULT_CHARSET;
	setName(name);
	setHeight(height);
	setStyle(style);
	if (pos >= str.length) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	wchar[] platform = str[pos++];

	if (pos >= str.length) return;
	wchar[] version2 = str[pos++];

	if (platform == "WINDOWS" && version2 == "1") {
		LOGFONT* newData = OS.IsUnicode
			? cast(LOGFONT*)new LOGFONTW()
			: cast(LOGFONT*)new LOGFONTA();
		//try {
			if (pos >= str.length) return;
			newData.lfHeight = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfWidth = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfEscapement = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfOrientation = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfWeight = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfItalic = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfUnderline = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfStrikeOut = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfCharSet = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfOutPrecision = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfClipPrecision = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfQuality = Strings!(wchar).toInt(str[pos++]);

			if (pos >= str.length) return;
			newData.lfPitchAndFamily = Strings!(wchar).toInt(str[pos++]);

		/*} catch (NumberFormatException e) {
			setName(name);
			setHeight(height);
			setStyle(style);
			return;
		}*/
		// FIXME: this will not do the Right Thing if the face name contains pipe characters.  -- andy
		// buglet introduced when I changed this from a lot of of find()s and substr()s to a split()
		TCHAR buffer = new TCHAR(0, str[pos], false);
		int length = OS.LF_FACESIZE <= buffer.length() ? OS.LF_FACESIZE - 1 : buffer.length(); // Math.min(OS.LF_FACESIZE - 1, buffer.length());
		if (OS.IsUnicode) {
			buffer.chars = (cast(LOGFONTW*)newData).lfFaceName;
		} else {
			buffer.bytes = (cast(LOGFONTA*)newData).lfFaceName;
		}
		data = newData;
	}
}

/**
 * Constructs a new font data given a font name,
 * the height of the desired font in points,
 * and a font style.
 *
 * @param name the name of the font (must not be null)
 * @param height the font height in points
 * @param style a bit or combination of NORMAL, BOLD, ITALIC
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - when the font name is null</li>
 *    <li>ERROR_INVALID_ARGUMENT - if the height is negative</li>
 * </ul>
 */
public this(wchar[] name, int height, int style) {
	if (name == null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
	data = OS.IsUnicode
		? cast(LOGFONT*)new LOGFONTW()
		: cast(LOGFONT*)new LOGFONTA();
	setName(name);
	setHeight(height);
	setStyle(style);
	// We set the charset field so that
	// wildcard searching will work properly
	// out of the box
	data.lfCharSet = OS.DEFAULT_CHARSET;
}

/**
 * Compares the argument to the receiver, and returns true
 * if they represent the <em>same</em> object using a class
 * specific comparison.
 *
 * @param object the object to compare with this object
 * @return <code>true</code> if the object is the same as this object and <code>false</code> otherwise
 *
 * @see #hashCode
 */
public bit equals (Object object) {
	if (object == this) return true;
// 	if (!(object instanceof FontData)) return false;
 	if (!(cast(FontData)object)) return false;
	FontData fd = cast(FontData)object;
	LOGFONT* lf = fd.data;
	return data.lfCharSet == lf.lfCharSet &&
		/*
		* This code is intentionally commented.  When creating
		* a FontData, lfHeight is not necessarily set.  Instead
		* we check the height field which is always set.
		*/
//		data.lfHeight == lf.lfHeight &&
		height == fd.height &&
		data.lfWidth == lf.lfWidth &&
		data.lfEscapement == lf.lfEscapement &&
		data.lfOrientation == lf.lfOrientation &&
		data.lfWeight == lf.lfWeight &&
		data.lfItalic == lf.lfItalic &&
		data.lfUnderline == lf.lfUnderline &&
		data.lfStrikeOut == lf.lfStrikeOut &&
		data.lfCharSet == lf.lfCharSet &&
		data.lfOutPrecision == lf.lfOutPrecision &&
		data.lfClipPrecision == lf.lfClipPrecision &&
		data.lfQuality == lf.lfQuality &&
		data.lfPitchAndFamily == lf.lfPitchAndFamily &&
		getName() == fd.getName();
}

int EnumLocalesProc(void* lpLocaleString) {

	/* Get the locale ID */
	int length = 8;
	TCHAR buffer = new TCHAR(0, length);
	int byteCount = length * TCHAR.sizeof;

	OS.MoveMemory(buffer, lpLocaleString, byteCount);
	int lcid = Strings!(wchar).toInt(buffer.toStringw(0, buffer.strlen ()), 16);

	/* Check the language */
	int size = OS.GetLocaleInfo(lcid, OS.LOCALE_SISO639LANGNAME, buffer, length);
	if (size <= 0 || lang != buffer.toStringw(0, size - 1)) return 1;

	/* Check the country */
	if (country != null) {
		size = OS.GetLocaleInfo(lcid, OS.LOCALE_SISO3166CTRYNAME, buffer, length);
		if (size <= 0 || country != buffer.toStringw(0, size - 1)) return 1;
	}

	/* Get the charset */
	size = OS.GetLocaleInfo(lcid, OS.LOCALE_IDEFAULTANSICODEPAGE, buffer, length);
	if (size <= 0) return 1;
	int cp = Strings!(wchar).toInt(buffer.toStringw(0, size - 1));
	int [] lpCs = new int[8];
	OS.TranslateCharsetInfo(cp, lpCs, OS.TCI_SRCCODEPAGE);
	data.lfCharSet = cast(byte)lpCs[0];
        return 0;
}

/**
 * Returns the height of the receiver in points.
 *
 * @return the height of this FontData
 *
 * @see #setHeight
 */
public int getHeight() {
	return height;
}

/**
 * Returns the locale of the receiver.
 * <p>
 * The locale determines which platform character set this
 * font is going to use. Widgets and graphics operations that
 * use this font will convert UNICODE strings to the platform
 * character set of the specified locale.
 * </p>
 * <p>
 * On platforms where there are multiple character sets for a
 * given language/country locale, the variant portion of the
 * locale will determine the character set.
 * </p>
 *
 * @return the <code>char[]</code> representing a Locale object
 * @since 3.0
 */
public wchar[] getLocale () {
	//StringBuffer buffer = new StringBuffer ();
	wchar[] buffer;
	wchar sep = '_';
	if (lang !== null) {
		buffer ~= lang;
		buffer ~= sep;
	}
	if (country !== null) {
		buffer ~= country;
		buffer ~= sep;
	}
	if (variant !== null) {
		buffer ~= variant;
	}

	if (buffer.length > 0) {
		if (buffer[buffer.length - 1] == sep) {
			buffer = buffer[0 .. buffer.length - 1];
		}
	}
	return buffer;
}

/**
 * Returns the name of the receiver.
 * On platforms that support font foundries, the return value will
 * be the foundry followed by a dash ("-") followed by the face name.
 *
 * @return the name of this <code>FontData</code>
 *
 * @see #setName
 */
public wchar[] getName() {
	wchar[] chars;
	if (OS.IsUnicode) {
		chars = (cast(LOGFONTW*)data).lfFaceName;
	} else {
		chars = new wchar[OS.LF_FACESIZE];
		char[] bytes = (cast(LOGFONTA*)data).lfFaceName;
		OS.MultiByteToWideChar (OS.CP_ACP, OS.MB_PRECOMPOSED, bytes, bytes.length, chars, chars.length);
	}

	// chop at terminating zero, C style
	int index = Strings!(wchar).find(chars, '\0');
	if (index == -1) index = chars.length;

	return chars[0 .. index];
}

/**
 * Returns the style of the receiver which is a bitwise OR of
 * one or more of the <code>DWT</code> constants NORMAL, BOLD
 * and ITALIC.
 *
 * @return the style of this <code>FontData</code>
 *
 * @see #setStyle
 */
public int getStyle() {
	int style = DWT.NORMAL;
	if (data.lfWeight == 700) style |= DWT.BOLD;
	if (data.lfItalic != 0) style |= DWT.ITALIC;
	return style;
}

/**
 * Returns an integer hash code for the receiver. Any two
 * objects which return <code>true</code> when passed to
 * <code>equals</code> must return the same value for this
 * method.
 *
 * @return the receiver's hash
 *
 * @see #equals
 */
public int hashCode () {
	return data.lfCharSet ^ height ^ data.lfWidth ^ data.lfEscapement ^
		data.lfOrientation ^ data.lfWeight ^ data.lfItalic ^data.lfUnderline ^
		data.lfStrikeOut ^ data.lfCharSet ^ data.lfOutPrecision ^
		data.lfClipPrecision ^ data.lfQuality ^ data.lfPitchAndFamily;
		//^ getName().hashCode();
}

/**
 * Sets the height of the receiver. The parameter is
 * specified in terms of points, where a point is one
 * seventy-second of an inch.
 *
 * @param height the height of the <code>FontData</code>
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_INVALID_ARGUMENT - if the height is negative</li>
 * </ul>
 *
 * @see #getHeight
 */
public void setHeight(int height) {
	if (height < 0) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
	this.height = height;
}

/**
 * Sets the locale of the receiver.
 * <p>
 * The locale determines which platform character set this
 * font is going to use. Widgets and graphics operations that
 * use this font will convert UNICODE strings to the platform
 * character set of the specified locale.
 * </p>
 * <p>
 * On platforms where there are multiple character sets for a
 * given language/country locale, the variant portion of the
 * locale will determine the character set.
 * </p>
 *
 * @param locale the <code>wchar[]</code> representing a Locale object
 * @see java.util.Locale#toString
 */
public void setLocale(wchar[] locale) {
	lang = country = variant = null;
	if (locale != null) {
		char sep = '_';
		int length = locale.length;
		int firstSep, secondSep;

		firstSep = Strings!(wchar).find(locale, sep);
		if (firstSep == -1) {
			firstSep = secondSep = length;
		} else {
			secondSep = Strings!(wchar).find(locale[firstSep + 1 .. locale.length], sep);
			if (secondSep == -1) secondSep = length;
		}
		if (firstSep > 0) lang = locale[0 .. firstSep];
		if (secondSep > firstSep + 1) country = locale[firstSep + 1 .. secondSep];
		if (length > secondSep + 1) variant = locale[secondSep + 1 .. locale.length];
	}
	if (lang == null) {
		data.lfCharSet = OS.DEFAULT_CHARSET;
	} else {
		void* pFunc = cast(void*)bind(&this.EnumLocalesProc);
		OS.EnumSystemLocales(pFunc, OS.LCID_SUPPORTED);
		unbind(&this.EnumLocalesProc);
		/*Callback callback = new Callback (this, "EnumLocalesProc", 1);
		int lpEnumLocalesProc = callback.getAddress ();
		OS.EnumSystemLocales(lpEnumLocalesProc, OS.LCID_SUPPORTED);
		callback.dispose ();*/
	}
}

/**
 * Sets the name of the receiver.
 * <p>
 * Some platforms support font foundries. On these platforms, the name
 * of the font specified in setName() may have one of the following forms:
 * <ol>
 * <li>a face name (for example, "courier")</li>
 * <li>a foundry followed by a dash ("-") followed by a face name (for example, "adobe-courier")</li>
 * </ol>
 * In either case, the name returned from getName() will include the
 * foundry.
 * </p>
 * <p>
 * On platforms that do not support font foundries, only the face name
 * (for example, "courier") is used in <code>setName()</code> and
 * <code>getName()</code>.
 * </p>
 *
 * @param name the name of the font data (must not be null)
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - when the font name is null</li>
 * </ul>
 *
 * @see #getName
 */
public void setName(wchar[] name) {
	if (name == null) DWT.error(DWT.ERROR_NULL_ARGUMENT);

	/* The field lfFaceName must be NULL terminated */
	TCHAR buffer = new TCHAR(0, name, true);
	int length = OS.LF_FACESIZE <= buffer.length() ? OS.LF_FACESIZE - 1 : buffer.length();  //Math.min(OS.LF_FACESIZE - 1, buffer.length());
	if (OS.IsUnicode) {
		wchar[] lfFaceName = (cast(LOGFONTW*)data).lfFaceName;
                lfFaceName = buffer.chars.dup;
		//for (int i = 0; i < lfFaceName.length; i++) lfFaceName[i] = 0;
		//System.arraycopy(buffer.chars, 0, lfFaceName, 0, length);
	} else {
		byte[] lfFaceName = (cast(LOGFONTA*)data).lfFaceName;
		for (int i = 0; i < lfFaceName.length; i++) lfFaceName[i] = 0;
		System.arraycopy(buffer.bytes, 0, lfFaceName, 0, length);
	}
}

/**
 * Sets the style of the receiver to the argument which must
 * be a bitwise OR of one or more of the <code>DWT</code>
 * constants NORMAL, BOLD and ITALIC.
 *
 * @param style the new style for this <code>FontData</code>
 *
 * @see #getStyle
 */
public void setStyle(int style) {
	if ((style & DWT.BOLD) == DWT.BOLD) {
		data.lfWeight = 700;
	} else {
		data.lfWeight = 0;
	}
	if ((style & DWT.ITALIC) == DWT.ITALIC) {
		data.lfItalic = 1;
	} else {
		data.lfItalic = 0;
	}
}

/**
 * Returns a string representation of the receiver which is suitable
 * for constructing an equivalent instance using the
 * <code>FontData(char[])</code> constructor.
 *
 * @return a string representation of the FontData
 *
 * @see FontData
 */
public char[] toString() {
	char[] fontDescription = "1|" ~ std.utf.toUTF8(getName()) ~ "|"
		~ .toString(getHeight()) ~ "|"
		~ .toString(getStyle()) ~ "|" ~ "WINDOWS|1|" 
		~ .toString(data.lfHeight) ~ "|"
		~ .toString(data.lfWidth) ~ "|" 
		~ .toString(data.lfEscapement) ~ "|" 
		~ .toString(data.lfOrientation) ~ "|"
		~ .toString(data.lfWeight) ~ "|" 
		~ .toString(data.lfItalic) ~ "|" 
		~ .toString(data.lfUnderline) ~ "|"
		~ .toString(data.lfStrikeOut) ~ "|"
		~ .toString(data.lfCharSet) ~ "|"
		~ .toString(data.lfOutPrecision) ~ "|"
		~ .toString(data.lfClipPrecision) ~ "|"
		~ .toString(data.lfQuality) ~ "|"
		~ .toString(data.lfPitchAndFamily) ~ "|"
		~ .toUTF8(getName());
        return fontDescription;
}

/**
 * Invokes platform specific functionality to allocate a new font data.
 * <p>
 * <b>IMPORTANT:</b> This method is <em>not</em> part of the public
 * API for <code>FontData</code>. It is marked public only so that
 * it can be shared within the packages provided by DWT. It is not
 * available on all platforms, and should never be called from
 * application code.
 * </p>
 *
 * @param data the <code>LOGFONT</code> for the font data
 */
public static FontData win32_new(LOGFONT* data, int height) {
	return new FontData(data, height);
}

}
