/*******************************************************************************
 * Copyright (c) 2000, 2003 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * DWT: SWT port for the D Language
 * 
 * Contributors:
 * 	The DWT team at http://www.dsource.org/projects/dwt
 * 
 *******************************************************************************/



module dwt.widgets.synchronizer;

/**
 * Instances of this class provide synchronization support
 * for displays. A default instance is created automatically
 * for each display, and this instance is sufficient for almost
 * all applications.
 * <p>
 * <b>IMPORTANT:</b> Typical application code <em>never</em>
 * needs to deal with this class. It is provided only to
 * allow applications which require non-standard
 * synchronization behavior to plug in the support they
 * require. <em>Subclasses which override the methods in 
 * this class must ensure that the superclass methods are
 * invoked in their implementations</em>
 * </p>
 *
 * @see Display#setSynchronizer
 */
public class Synchronizer {

private import dwt.dwt;

private import dwt.internal.compatibility;

private import dwt.util.runnable;
private import dwt.util.system;
private import dwt.util.thread;
private import dwt.util.throwable;

private import dwt.widgets.display;
private import dwt.widgets.runnablelock;

	Display display;
	int messageCount;
	RunnableLock [] messages;
	Object messageLock;
	Thread syncThread;

public this (Display display) {
	this.display = display;
	messageLock = new Object ();
}
	
public void addLast(RunnableLock lock) {
	synchronized (messageLock) {
		if (messages == null) messages = new RunnableLock [4];
		if (messageCount == messages.length) {
			RunnableLock[] newMessages = new RunnableLock [messageCount + 4];
			System.arraycopy (messages, 0, newMessages, 0, messageCount);
			messages = newMessages;
		}
		messages [messageCount++] = lock;
	}
}

/**
 * Causes the <code>run()</code> method of the runnable to
 * be invoked by the user-interface thread at the next 
 * reasonable opportunity. The caller of this method continues 
 * to run in parallel, and is not notified when the
 * runnable has completed.
 *
 * @param runnable code to run on the user-interface thread.
 *
 * @see #syncExec
 */
//protected 
public void asyncExec (Runnable runnable) {
	if (runnable != null) addLast (new RunnableLock (runnable));
	display.wake ();
}

public int getMessageCount () {
	return messageCount;
}

public void releaseSynchronizer () {
	display = null;
	messages = null;
	messageLock = null;
	syncThread = null;
}

public RunnableLock removeFirst () {
	synchronized (messageLock) {
		if (messageCount == 0) return null;
		RunnableLock lock = messages [0];
		System.arraycopy (messages, 1, messages, 0, --messageCount);
		messages [messageCount] = null;
		if (messageCount == 0) messages = null;
		return lock;
	}
}

public bool runAsyncMessages () {
	if (messageCount == 0) return false;
	RunnableLock lock = removeFirst ();
	if (lock == null) return true;
	synchronized (lock) {
		syncThread = lock.thread;
		try {
			lock.run ();
		} catch(Throwable t) {
			lock.throwable = t;
			DWT.error(DWT.ERROR_FAILED_EXEC, true);
		} finally {
			syncThread = null;
// TODO: What's D equivalent ??			lock.notifyAll ();
		}
	}
	return true;
}


/**
 * Causes the <code>run()</code> method of the runnable to
 * be invoked by the user-interface thread at the next 
 * reasonable opportunity. The thread which calls this method
 * is suspended until the runnable completes.
 *
 * @param runnable code to run on the user-interface thread.
 *
 * @exception SWTException <ul>
 *    <li>ERROR_FAILED_EXEC - if an exception occured when executing the runnable</li>
 * </ul>
 *
 * @see #asyncExec
 */
//protected void syncExec(Runnable runnable) {
public void syncExec(Runnable runnable) {
	if (display.isValidThread ()) {
		if (runnable != null) runnable.run ();
		return;
	}
	if (runnable == null) {
		display.wake ();
		return;
	}
	RunnableLock lock = new RunnableLock (runnable);
	/*
	 * Only remember the syncThread for syncExec.
	 */
	lock.thread = cast(Thread)Thread.currentThread();
	synchronized (lock) {
		addLast (lock);
		display.wake ();
		bit interrupted = false;
		while (!lock.done ()) {
//			try {
				lock;
//	TODO: What's D equivalent??			lock.wait();
//			} catch (InterruptedException e) {
//				interrupted = true;
//			}
		}
		if (interrupted) {
// What's D equivalent???			Compatibility.interrupt();
//			Compatibility;
		}
		if (lock.throwable != null) {
			DWT.error(DWT.ERROR_FAILED_EXEC, lock.throwable == null ? true : false);
		}
	}
}

}
